(* ===================================================================

  This file provides:

  (1) ML data-types INSTR, VLDSCRIPT, and PROG to represent L0 
      sentences as trees.

  (2) The coresponding algebra-type and fold for INSTR

  author: wishnu prasetya

=================================================================== *)

local 

open Parse HolKernel boolLib  ;
open Susp ; (* support for forcing lazy evaluation *)

in

datatype ('ident       (* identifier *)
         ,'expr        (* program's expression *)
         ,'assert      (* assertion  *)
         ,'term)       (* the original HOL term, to be stored lazily;
                          mostly used for error reporting *)
         INSTR = 
       PRINT of { pexpr : 'expr
                , pself : 'term susp
                }
     | ASG   of { alhs : 'expr
                , arhs : 'expr 
                , aself: 'term susp
                }
     | SEQ   of { sbody : (('ident, 'expr, 'assert, 'term) INSTR) list
                , sself : 'term susp
                }
       (* NOTE:
          "skip" is not represented explicitly, but by using
          an empty SEQ.
       *)
     | CALL  of { clhs : 'expr         (* target var *)
              , cpname : 'ident        (* called prg name *)
              , cparams  : 'expr list  (* all parameters, in the right order *)
              , crparams : 'ident list (* pass by reference params *)
              , cvparams : 'expr list  (* pass by value params *)
              , cself : 'term susp
              }
       (* NOTE:
          using the same representation for /@ and /@=;
          in the case of /@, it will be treated as "skip /@= P(...)"
        *)
     | IF    of { iguard : 'expr 
              , ithen : ('ident, 'expr, 'assert, 'term) INSTR
              , ielse : ('ident, 'expr, 'assert, 'term) INSTR
              , iself : 'term susp
              }
     | LOCAL of { lvars : ('ident * 'expr) list
              , lbody : ('ident, 'expr, 'assert, 'term) INSTR
              , lself : 'term susp
              }
     | WHILE of { wguard : 'expr
              , winv : 'expr option
              , wbody : ('ident, 'expr, 'assert, 'term) INSTR
              , wself : 'term susp
              }
     | ASSERT of { aexpr : 'expr
                 , aself : 'term susp
                 }
  ;

type ('ident,'expr,'assert,'instr,'rettype) PROG
     =
     { pname : 'ident
     , pparams  : 'expr  list  (* all parameters, in the right order *)
     , prparams : 'ident list
     , pvparams : 'ident list
     , ppre     : 'assert
     , ppost    : 'assert
     , pbody    : 'instr
     , preturn  : 'rettype
     }
    ;

(* ===================================================================

  INSTR algebra

=================================================================== *)        

type ('ident,'expr,'assert,'result,'term) alg_INSTR
     =
     { fprint : 'term susp -> 'expr -> 'result

     , fasg : 'term susp -> 'expr -> 'expr -> 'result

     , fseq : 'term susp -> 'result list -> 'result

     , fcall: 'term susp 
              ->
              'expr -> 'ident -> 'expr list -> 'ident list -> 'expr list -> 'result

     , fif :  'term susp -> 'expr -> 'result -> 'result -> 'result

     , flocal: 'term susp 
               ->  
               ('ident * 'expr) list -> 'result -> 'result

     , fwhile: 'term susp 
               -> 
               'expr -> 'expr option -> 'result -> 'result

     , fassert: 'term susp -> 'expr -> 'result

     } ;

(* ===================================================================

  INSTR fold

=================================================================== *)        

fun fold_INSTR 
      
   (alg : ('ident,'expr,'assert,'result,'term) alg_INSTR)

   (instr : ('ident,'expr,'assert,'term) INSTR)

    =
    let 

    fun fold i = case i of

        PRINT e   => (#fprint alg) (#pself e) (#pexpr e)
    
     |  ASG a     => (#fasg alg) (#aself a) (#alhs a) (#arhs a)

     |  SEQ body  => (#fseq alg) (#sself body) (map fold (#sbody body))

     |  CALL c    => (#fcall alg)(#cself c)
                                 (#clhs c)
                                 (#cpname c)
                                 (#cparams c)
                                 (#crparams c)
                                 (#cvparams c)
      
     | IF c       => (#fif alg)  (#iself c)
                                 (#iguard c)
                                 (fold (#ithen c))
                                 (fold (#ielse c))

     | LOCAL b    => (#flocal alg) (#lself b)
                                   (#lvars b)
                                   (fold (#lbody b))

     | WHILE w    => (#fwhile alg) (#wself w)
                                   (#wguard w)
                                   (#winv   w)
                                   (fold (#wbody w))

     | ASSERT a   => (#fassert alg) (#aself a) (#aexpr a)

    in
    fold instr
    end ;


(* ===================================================================

  A concrete instance of INSTR and alg_INSTR

=================================================================== *)        

type PredTransf    = term->term ; 
type INSTR_        = (term,term,term,term) INSTR ;
type 'r alg_INSTR_ = (term,term,term,'r,term) alg_INSTR ;
type wpalg_INSTR_  = (term,term,term,PredTransf,term) alg_INSTR ;

end ;