(* ===================================================================

  This file contains the (algebraic) specification of WP-reduction 
  for L0min.

  author: wishnu prasetya

=================================================================== *)

local 

open Parse HolKernel boolLib bossLib ;

open Susp Option Err 
open L0Theory L0_emb_mldata L0_emb_basic 
open Logic_Util Logic Logic_gen ;

infixr 5 AND OR ;
infixr 4 IMP ;
infixr   //  ;
infix    <-  ;

fun matchTerm t1 t2 = SOME (fst (match_term t1 t2)) 
                      handle _ => NONE ;

fun snd (_,y) = y ;
fun fst (x,_) = x ;

in 

(* ===================================================================

   Here is a standard partial correctness WP-LOGIC for L0min 

=================================================================== *)


structure L0min_logic : Logic =
   struct

   type Lstate   = VC list
   type Loption = { ErrorEchoOn : bool }

   (* a function to guess invariant *)
   val guessInv  = ref noguessInv

   (* a function to retrieve the (entire) specification of a program *)
   val getSpec   = ref simple_getSpec

   val option   = ref { ErrorEchoOn = true }

   val state              = ref ([] : Lstate) 
   fun initializeState () = (state := [])
   fun add_VCs vcs        = (state := vcs @ !state)
   fun get_VCs ()         = !state

   fun isErrorEchoOn ()   = #ErrorEchoOn (!option)

   val wpAlgebra  = 

    {

    (* ===================================================================

       The print rule:  WP (print e) Q = Q

    =================================================================== *)

     fprint = let
              fun wpprint self e Q = Q
              in
              wpprint end

    (* ===================================================================

       The assignment rule, as usual:  WP (x /:= e) Q = Q[e/x]

       However, we have to make sure that occurrence of "old x" in Q
       should not be substituted, as it refers to the old value of x.
       So:

            WP (x /:= e) Q = Q[@x/old x]
                              [e/x]
                              [old x/@x]

    =================================================================== *)

    ,fasg = let
            fun wpasg self x e Q 
                = 
                let
                val fresh     = mkFreshVars [x]
                in
                Q <- (fresh//[mk_old x])
                  <- ([e]//[x])
                  <- ([mk_old x]//fresh)
                end
            in 
            wpasg end

    (* ===================================================================

       The seq rule:  WP (i;j) Q = WP i (WP j Q)

    =================================================================== *)
    
    ,fseq = let
            fun wpseq self wps Q =  List.foldr (op o) I wps Q
            in
            wpseq end

    (* ===================================================================

       The program-call rule:  

       Assuming we have the following black-box spec for a program S:

               {P} S(r,v) {Q}

       Assuming r is passed-by-ref and v by-val.

       Note that if v is referred in Q, it means the initial value of v;
       that is, its value before entering S.

       Then WP (x /@= S(s,e)) {R}
            =
            P[r/old r,v/old v][s/r,e/v]  /\  

           ( Q [@s/r]                  
               [v/old v]
               [e/v]
               [s/old @s]
             ==>
             R [ret/x][@z/old s][@s/s][old s/@z] )
   
       where @name represents freshly generated name.
 
    =================================================================== *)

   ,fcall =  
      let
      fun wpcall self x pname allparams s e R
           =
           let 
           val (pname'
                    ,_
                    ,r
                    ,v
                    ,(P,Q,_,_)) 
                     =
                       (xdestProgDef
                           o concl
                           o TY_INST_prog_def allparams
                           o !getSpec 
                           o fst 
                           o dest_const) pname

           val fresh1 = mkFreshVars s
           val fresh2 = mkFreshVars s

           fun app_old t = let
                           val ty  = type_of t
                           val old = mk_const ("old", ty --> ty)
                           in
                           mk_comb (old,t)
                           end

           val ret = mk_const ("ret", type_of x)

           val wp = (P <- (r @ v // mk_olds (r @ v)) <- (s @ e // r @ v))
                    AND
                    ( (Q <- (fresh1 //r)                  
                         <- (v // mk_olds v)
                         <- (e//v)
                         <- (s// mk_olds fresh1))
          
                      IMP
 
                      (R <- ([ret]  // [x]) 
                         <- (fresh2 // mk_olds s)
                         <- (fresh1 // s)
                         <- (mk_olds s // fresh2))
                    )

           in
           ( if pname=pname' then wp
             else raise 
                  L0Exn 
                     { errorCode  = SOME "fcall in L0minLogic"
                     , errorWhere = SOME (force self)
                     , errorInside = NONE
                     , errorWhy   = SOME "the stored spec is either of another program, or there is a problem in type-instantiating it" } 
           )
           handle  
              _ =>
              raise  L0Exn
                     { errorCode  = SOME "program call"
                     , errorWhere = SOME (force self)
                     , errorInside = NONE
                     , errorWhy   = SOME "somehow fails to properly instantiate the prog's spec" }
           end
    in
    wpcall end 

    (* ===================================================================

       The cond rule:  

       WP (if g then i1 else i2) Q  =  if g then WP i1 Q else WP i2 Q

    =================================================================== *)

    ,fif = let
           fun wpif self g wpi1 wpi2 Q
                    = 
                    mk_cond (g, wpi1 Q, wpi2 Q)
           in
           wpif end

    (* ===================================================================

       The block rule:  

       WP (let x=e in body) Q  =  (WP body Q[@x/x])[x/old x][e/x][x/@x]

       also, the [@x/x] has to be applied to VCs just before the
             recursion into body, and then [e/old x][x/@x] is
             applied to VCs after the recursion to body.

       Note that x occurs in Q refer to the x before the block is
       entered.
    =================================================================== *)

   ,flocal =  let
              fun wplocal self locdefs wpbody Q = 
                  let
                  val (x,e) = unzip locdefs
                  val fresh = mkFreshVars (List.map fst locdefs)
                  val w     = ref mk_T
                  in
                  ( state := List.map (fn v => { VClabel = #VClabel v
                                               , VC = #VC v <- (fresh//x)
                                               } 
                                       )
                                      (!state)
                  ; w   :=  ((wpbody (Q <- (fresh//x)))
                               <- (x // mk_olds x)
                               <- (e//x) 
                               <- (x//fresh))
                  ; state := List.map (fn v => { VClabel = #VClabel v
                                               , VC = #VC v <- (e//mk_olds x)
                                                          <- (x//fresh)   
                                               }
                                      )
                                      (!state)
                  ; !w
                  )
                  end
              in
              wplocal end

    (* ===================================================================

       The while rule:  (partial correctness!)

       WP (while g using i wdo body) Q  =  i
          
           generating these vcs:

           i /\ g   ==> WP body i
           i /\ ~g  ==> Q

    =================================================================== *)

   ,fwhile =  let
              fun wpwhile self g i wpbody Q
                     = 
                     let
                     val inv = case i of
                                 SOME i_ => i_
                               | NONE    => !guessInv (force self) Q

                     val vc1 = case i of
                                 SOME i_ => { VClabel = "IC"
                                            , VC = (inv AND g) IMP rand (wpbody inv)
                                            (* adding "rand", else we get double
                                               i's in the VC *)
                                            }
                               | NONE => { VClabel = "IC"
                                            , VC = (inv AND g) IMP (wpbody inv)
                                            }

                     val vc2 = { VClabel = "EC",
                                 VC = (inv AND NOT g) IMP Q
                               }
                     in
                     ( add_VCs [vc1,vc2] ; inv )
                     end 
              in
              wpwhile end

    (* ===================================================================

       The assert rule:  

       WP (assert a) Q  =  p /\ Q

    =================================================================== *)

   ,fassert = let
              fun wpassert self a Q = a AND Q
              in 
              wpassert end

   } ;

  end

(* ===================================================================

  Now generating the coresponding VCG-utilities

=================================================================== *)


structure L0min_vcg = build_VCG(L0min_logic) ;


end ;
