(* =============================================================

   The specification of L0min syntax checker

============================================================= *)

local

open Parse HolKernel boolLib integerTheory
open Susp Err 
open L0_emb_mldata 
     SChecker SChecker_gen
     Test_SChecker_gen
     L0min_Syntax 

fun matchTerm t1 t2 = SOME (fst (match_term t1 t2)) 
                      handle _ => NONE ;

fun snd (_,y) = y ;
fun fst (x,_) = x ;

fun distinct [] = true
  | distinct (x::s) = (not o isSome o List.find (fn y => x=y)) s
                      andalso
                      distinct s

fun elem x []     = false
  | elem x (y::s) = x=y orelse elem x s

fun is_ret t = (fst o dest_const) t = "ret"
               handle 
               _ => false

fun is_skip t = (fst o dest_const) t = "skip"
               handle 
               _ => false

in

datatype L0minMsg = L0minError   of L0Error
                  | L0minWarning of L0Warning

structure SCspec : SChecker =

   struct

   type SCstate    = L0minMsg list
   type SCoption   = unit (* no option available atm *)
   type SCmsg      = L0minMsg

   val state       = ref []
   val option      = ref ()

   fun initializeState ()   = (state := [])
   fun add_SCmsgs msgs      = (state := msgs @ !state)
   fun get_SCmsgs ()        = !state
   fun print_SCmsg (L0minError e)   = printL0Error   e
     | print_SCmsg (L0minWarning e) = printL0Warning e

   val L0Error2SCmsg   = L0minError
   val L0Warning2SCmsg = L0minWarning 

   fun check_expr t 
       =
       (if isSimpleExpr t then [] else [])
        handle
        L0Exn e => [L0Error2SCmsg e]

   fun check_assertion t
       =
      (if isAssertion t then [] else [])
       handle
       L0Exn e => [L0Error2SCmsg e]

   fun check_assertion_ret t
       =
      (if isAssertion_ret t then [] else [])
       handle
       L0Exn e => [L0Error2SCmsg e]

   fun check_allowedtype t 
       = 
       let
       val ty = type_of t
       in
       if elem ty [ bool
                  , Type `:int`
                  , Type `:string`
                  , Type `:void` ]
       then []
       else [ L0Error2SCmsg
                         { errorCode  = SOME "check_allowedtype"
                           , errorInside = NONE
                           , errorWhere = SOME t
                           , errorWhy   = SOME "its type is not allowed in L0mins"  }
            ] 
       end

   fun check_actual_param t =  
       let
       fun isRefPar t =  isSome (matchTerm (Term `REF jkjakjh`) t)
       in
       if isRefPar t 
       then let
            val t_ = rand t
            fun check_ident t 
                 = 
                 if is_var t 
                 then []
                 else [ L0Error2SCmsg
                           { errorCode  = SOME "check_actual_param"
                           , errorWhere = SOME t
                           , errorInside = NONE
                           , errorWhy   = SOME "not allowed as pass-by-reference parameter"  }
                      ]
            in
            check_allowedtype t_ @ check_ident t_
            end
       else check_allowedtype t
       end

   val scAlgebra   =

     { fprint  = 
         let
         fun checkprint _ e = add_SCmsgs (check_expr e)
         in 
         checkprint end

     , fasg =
         let
         fun checkasg self v e 
                = 
                let
                fun check_ident v 
                    =
                    if is_var v orelse is_ret v
                    then ()
                    else add_SCmsgs 
                         [ L0Error2SCmsg
                           { errorCode   = SOME "L0min syntax"
                           , errorWhere  = SOME v 
                           , errorInside = SOME (force self)
                           , errorWhy    = SOME "the target should be an identifier"  } ]
                in
                (add_SCmsgs (check_expr e) ;
                 check_ident v )
                end
         in 
         checkasg end
         
     , fseq    = 
        let
        fun checkseq _ rs = (rs; ())
        in
        checkseq end

     , fcall   =
        let
        fun checkcall self v pname pars rpars vpars 
            =
            let
            fun check_target x 
                    =
                    if is_var x orelse is_skip x 
                    then ()
                    else add_SCmsgs 
                         [ L0Error2SCmsg
                           { errorCode   = SOME "L0min syntax"
                           , errorWhere  = SOME x
                           , errorInside = SOME (force self)
                           , errorWhy   = SOME "target should be an identifier"  } ]
            fun check_pname x 
                    =
                    if is_const x 
                    then ()
                    else add_SCmsgs 
                         [ L0Error2SCmsg
                           { errorCode   = SOME "L0min syntax"
                           , errorWhere  = SOME x
                           , errorInside = SOME (force self)
                           , errorWhy   = SOME "program-name should be a (HOL) constant"  } ]
            fun check_alias rpars 
                = 
                if distinct (List.filter is_var rpars)
                then ()
                else add_SCmsgs 
                         [ L0Error2SCmsg
                           { errorCode   = SOME "L0min syntax"
                           , errorWhere  = NONE
                           , errorInside = SOME (force self)
                           , errorWhy    = SOME "some pass-by-reference parameters are duplicate (alias problem)"  } ]
            in
            ( check_target v
            ; check_pname pname
            ; check_alias rpars
            ; add_SCmsgs (List.concat (List.map check_actual_param pars))
            ; ()
            )
            end
        in
        checkcall end

     , fif     = 
        let
        fun checkif _ g r1 r2 = (add_SCmsgs (check_expr g); r1; r2)
        in
        checkif end

     , flocal  = 
         let
         fun checkloc self decls rbody 
             =
             let
             val bvars = List.map fst decls
             val rhss  = List.map snd decls

             fun check_distinct bvars =
                if distinct (List.filter is_var bvars)
                then ()
                else add_SCmsgs 
                         [ L0Error2SCmsg
                           { errorCode   = SOME "L0min syntax"
                           , errorWhere  = NONE
                           , errorInside = SOME (force self)
                           , errorWhy   = SOME "some local variables are declared double within the same scope"  } ]

             in
             ( check_distinct bvars
             ; List.map check_allowedtype bvars
             ; add_SCmsgs (List.concat (List.map check_expr rhss))
             ; rbody )
             end
         in
         checkloc end

     , fwhile  = 
          let
          fun checkwhile _ g i rbody 
              =
              ( add_SCmsgs (check_expr g)
              ; (case i of
                     SOME j => add_SCmsgs (check_assertion j)
                   | _      => () )
              ; rbody )
          in
          checkwhile end

     , fassert = 
          let
          fun checkassert _ a = add_SCmsgs (check_assertion a)
          in 
          checkassert end
     } 


   end (* of struc *)

(* ===================================================================

  Now generating the coresponding SC-utilities

=================================================================== *)

structure SC  = build_SC(SCspec) ;
structure TSC = build_TSC(SCspec) ;
end