(* =============================================================

   L0min is a simplistic instance of L0. It is L0, in which
   types, expressions, and assertions are restricted as follows:

        type --> void | bool | int | string

        Expr   --> T | F | int-literal | string-lit | void | Ident
                 | Expr binop Expr
                 | unop Expr

        binop --> +,-,*,/,= 
                 ,>,>=,<,<=
                 ,/\,\/,==>

        unop --> ~ (bool,int)

        // assertion

        A   --> T | F | int-literal | string-lit | void | Ident
              | old Ident
              | A binop A
              | unop A
              | !bvars. A
              | ?bvars. A

        // assertion-ret: the same as assertion, but allows "ret"

   This file contains supporting syntax-utilities for L0min

============================================================= *)

local

open Parse HolKernel boolLib intLib stringLib stringSyntax
open Err L0Theory

fun snd (_,y) = y ;
fun fst (x,_) = x ;

fun matchTerm t1 t2 = SOME (fst (match_term t1 t2)) 
                      handle _ => NONE ;

fun elem x []     = false
  | elem x (y::s) = x=y orelse elem x s

val get_binop = rator o rator

fun args_binop t = (rand (rator t), rand t)

val qAll = Term `$!`

val qExists = Term `$?`

in

val L0min_binops = List.map Term
                   [ `($+): int->int->int`
                   , `($-): int->int->int`
                   , `($*): int->int->int`
                   , `($/): int->int->int`
                   , `($=): int->int->bool`
                   , `($=): bool->bool->bool`
                   , `($=): void->void->bool`
                   , `($=): string->string->bool`
                   , `($>): int->int->bool`
                   , `($>=): int->int->bool`
                   , `($<): int->int->bool`
                   , `($<=): int->int->bool`
                   , `($/\): bool->bool->bool`
                   , `($\/): bool->bool->bool`
                   , `($==>): bool->bool->bool`
                   ] ;
					     
val L0min_unops = [ Term `($~):bool->bool`
                  , Term `($~):int->int`   ] ;

fun is_L0min_binop t = elem (get_binop t) L0min_binops
                      handle
                      _ => false 

fun is_L0min_unop t = elem (rator t) L0min_unops
                     handle
                     _ => false 

fun isSimpleExpr t =
       if  t = Term `T`     orelse
           t = Term `F`     orelse
           t = Term `void`  orelse
           intSyntax.is_int_literal t    orelse
           is_string_literal t orelse
           is_var t
       then true
       else if is_L0min_unop  t  then isSimpleExpr (rand t)
       else if is_L0min_binop t  then
               let
               val (t1,t2) = args_binop t
               in
               isSimpleExpr t1 andalso isSimpleExpr t2
               end
       else raise
            L0Exn { errorCode   = SOME "isSimpleExpr"
                  , errorInside = NONE
                  , errorWhere  = SOME t
                  , errorWhy    = SOME "not allowed as a L0min simple-expr"  } 
      ;

fun isAssertion_ withRet t =
    let
    fun check t =   
       if  t = Term `T`     orelse
           t = Term `F`     orelse
           t = Term `void`  orelse
           intSyntax.is_int_literal t orelse
           is_string_literal t orelse
           is_var t
       then true       
       (* to handle "old v" *)
       else if isSome (matchTerm (Term `old jhgajak`) t)
            then is_var (rand t)  (* v shound be a var *) 
       (* to handle the "ret" keyword *)
       else if withRet andalso is_const t then (fst o dest_const) t = "ret"
       else if is_L0min_unop t   then check (rand t)
       else if is_L0min_binop t  then
               let
               val (t1,t2) = args_binop t
               in
               check t1 andalso check t2
               end
       else if is_forall t then (check o snd o strip_forall) t
       else if is_exists t then (check o snd o strip_exists) t
       else raise
            L0Exn { errorCode  = SOME "isAssertion_"
                  , errorInside = NONE
                  , errorWhere = SOME t
                  , errorWhy   = SOME "not allowed as an L0min assertion"  } 
     in
     check t
     end
     ;

val isAssertion     = isAssertion_ false
val isAssertion_ret = isAssertion_ true

(* -- a datatye as a structural representation of both simple expressions 
      and assertions;
      it will be used as intermediate representation during translation
      e.g. to executables.
*)

datatype L0minExpr =
         Const of term
       | Ident of term
       | Binop of ( term           (* the operator *)
                  * L0minExpr
                  * L0minExpr)
       | Unop  of ( term           (* the operator *)
                  * L0minExpr )
       | Old   of L0minExpr        (* "old" expression *)
       | Quant of ( term
                  * term list      (* the bvars *)
                  * L0minExpr )

(*-- functions to build L0minExpr from HOL-term 

     build_simple_L0minExpr takes a HOL-term containing a simple L0min
     expression, and construct the corresponding L0minExpr representation.     
     
     build_L0minExpr takes a HOL-term containing an L0min assertion, and construct 
     the corresponding L0minExpr representation.     

     (!) Both functions ASSUME that the input-term has been 
     syntax-checked

*)

fun build_simple_L0minExpr t =
    let
    fun build t =
        if is_const t  (* will also take care of "void" *)
           orelse
           intSyntax.is_int_literal t 
           orelse
           is_string_literal t
        then Const t
        else if is_var t    then Ident t
        else if is_L0min_unop t   then Unop (rator t, build (rand t))
        else if is_L0min_binop t  then
               let
               val (t1,t2) = args_binop t
               in
               Binop (rator (rator t), build t1, build t2)
               end
       else raise
            L0Exn { errorCode  = SOME "build_simple_L0minExpr"
                  , errorInside = NONE
                  , errorWhere = SOME t
                  , errorWhy   = SOME "not an L0min simple-expr; it should be syntax-checked first!"  }
    in
    build t
    end  
     
fun build_L0minExpr t =
    let
    fun build t =
       if is_const t  (* will also take care of "void" and "ret" *)
           orelse
           intSyntax.is_int_literal t 
           orelse
           is_string_literal t
       then Const t
       (* to handle "old" keyword *)
       else if isSome (matchTerm (Term `old jhgajak`) t)
            then Old (build (rand t))
       else if is_var t    then Ident t
       else if is_L0min_unop t   then Unop (rator t, build (rand t))
       else if is_L0min_binop t  then
               let
               val (t1,t2) = args_binop t
               in
               Binop (rator (rator t), build t1, build t2)
               end
       else if is_forall t then 
               let
               val (bvars,body) = strip_forall t
               in
               Quant (qAll, bvars, build body)
               end
       else if is_exists t then 
               let
               val (bvars,body) = strip_exists t
               in
               Quant (qExists, bvars, build body)
               end
       else raise
            L0Exn { errorCode  = SOME " build_L0minExpr"
                  , errorInside = NONE
                  , errorWhere = SOME t
                  , errorWhy   = SOME "not an L0min assertion; it should be syntax-checked first!"  }
    in
    build t
    end  

end